require "Window"
 
-----------------------------------------------------------------------------------------------
-- MTM Module Definition
-----------------------------------------------------------------------------------------------
local MTM = {}
local MRF = nil
 
-----------------------------------------------------------------------------------------------
-- Constants
-----------------------------------------------------------------------------------------------
-- e.g. local kiExampleVariableMax = 999
 
-----------------------------------------------------------------------------------------------
-- Initialization
-----------------------------------------------------------------------------------------------
function MTM:new(o)
    o = o or {}
    setmetatable(o, self)
    self.__index = self 

    -- initialize variables here

    return o
end

function MTM:Init()
	local bHasConfigureFunction = false
	local strConfigureButtonText = "MischhTargetingModule"
	local tDependencies = {"MischhRaidFrames"}
    Apollo.RegisterAddon(self, bHasConfigureFunction, strConfigureButtonText, tDependencies)
end
 

-----------------------------------------------------------------------------------------------
-- MTM OnLoad
-----------------------------------------------------------------------------------------------
function MTM:OnLoad()
    -- load our form file
	MTM = self
	self.xmlDoc = XmlDoc.CreateFromFile("MTM_xml.xml")
	self.xmlDoc:RegisterCallback("OnDocLoaded", self)
	
	--scince we set the dependency, we can assume MRF is loaded (aka has :OnLoad executed already)
	MRF = Apollo.GetAddon("MischhRaidFrames")
end

function MTM:LoadForm(name, parent, handler)
	return Apollo.LoadForm(self.xmlDoc, name, parent, handler)
end

-----------------------------------------------------------------------------------------------
-- MTM OnDocLoaded
-----------------------------------------------------------------------------------------------
do
	local load = {}
	local loaded = false
	
	local function exec(f, t)
		if not t then
			f()
		elseif type(t[f]) == "function" then
			t[f](t)
		end
	end
	
	function MTM:OnDocLoaded()
		loaded = true
		
		for f,t in pairs(load) do
			exec(f,t)
		end 
	end
	
	function MTM:OnceDocLoaded(func, tbl)
		if not func then return end
			
		if not loaded then
			if type(tbl) == "table" then
				load[func] = tbl --if tbl is actually a table, we assume func is a correct key for it.
			elseif type(func) == "function" then
				load[func] = false
			end --all other situations are useless.
		else
			if type(tbl) == "table" then
				if type(tbl[func]) == "function" then
					tbl[func](tbl)
				end
			elseif type(func) == "function" then
				func()
			end
		end
	end
end
-----------------------------------------------------------------------------------------------
-- MTM Core Functionality
-----------------------------------------------------------------------------------------------

MTM:OnceDocLoaded(function() MRF:OnceDocLoaded("Debug_OnLoadedDocs", MTM) end)
function MTM:Debug_OnLoadedDocs()
	local bSuccess, err = pcall(function()
		MTM:OnLoadedDocs()
	end)
	if not bSuccess then
		TEST = err
		Print("Error while loading MTM: "..err)
	end
end
function MTM:OnLoadedDocs()
	local frame = MTM:LoadForm("Window", "InWorldHudStratum")
	local targetForm = MTM:LoadForm("Target", frame)
	local ModOpt = MRF:GetOption(nil, "TargetingModule")
	local target = GameLib.GetTargetUnit() --for selection and drawing
	
	local function disabledByKey() return true end --overwritten in 'do --key disablers'
	local function scoreUnit(tUnit, pVec, pFac) return 1 end --overwritten in 'do --score units'
	local combatDisabler = false --overwritten in 'do --out-of-combat disabling'
	
	local timer = nil
	local function activated(tbl)
		tbl = tbl or MRF:GetOption(ModOpt, "classes"):Get()
		if not tbl then return false end
		local player = GameLib.GetPlayerUnit()
		if not player then
			timer = ApolloTimer.Create(3, false, "ForceUpdate", MRF:GetOption(ModOpt, "classes"))
			return false
		end
		local i = player:GetClassId()
		return tbl[i] and true or false
	end
	
	
	do --activation
		local classOpt = MRF:GetOption(ModOpt, "classes")
		MRF:GatherUpdates(classOpt)
		classOpt:OnUpdate(function(tbl) 
			if not tbl then
				classOpt:Set({	
					[GameLib.CodeEnumClass.Warrior] = false,
					[GameLib.CodeEnumClass.Engineer] = false,
					[GameLib.CodeEnumClass.Esper] = true,
					[GameLib.CodeEnumClass.Medic] = false,
					[GameLib.CodeEnumClass.Stalker] = false,
					[GameLib.CodeEnumClass.Spellslinger] = false,
				})
			else
				if activated(tbl) then
					Apollo.RegisterEventHandler("FrameCount", "OnFrameCount", MTM)
				else
					Apollo.RemoveEventHandler("FrameCount", MTM)
				end
			end
		end)
	end
	
	do --key disablers
		local keyOpt = MRF:GetOption(ModOpt, "disableKeys") --{["alt"] = bool, ["shift"] = bool, ["ctrl"] = bool, ["caps"] = bool, ["capsOnOff"] = bool} --On = true; Off = false
		MRF:GatherUpdates(keyOpt)
		
		--scince I'm too lazy to just write down all combinations, we compile a new function from scratch.
		
		keyOpt:OnUpdate(function(tbl)
			if not tbl then
				keyOpt:Set({alt=false, shift=true, ctrl=false, caps=true, capsOnOff=true})
			else
				local content = {}
				if tbl.alt then table.insert(content, "alt()") end
				if tbl.shift then table.insert(content, "shift()") end
				if tbl.ctrl then table.insert(content, "ctrl()") end
				if tbl.caps then table.insert(content, "caps()"..(tbl.capsOnOff and "" or " == false")) end
				local f = "function() return "..table.concat(content, " or ").." end" -- <- this is what we want compiled into 'disabledByKey()'
				--example at this point: f = "function() return alt() or ctrl() or caps() == false end"
				--only need to feed alt(), shift(), ... to it:
				f = "return function(alt, shift, ctrl, caps) return "..f.." end" 
				--now: if we throw this through loadstring() we get a function, which returns what we want to execute.
				disabledByKey = loadstring(f)()(Apollo.IsAltKeyDown, Apollo.IsShiftKeyDown, Apollo.IsControlKeyDown, Apollo.IsCapsLockOn)
			end
		end)
	end
	
	do --out-of-combat disabling
		local oocOpt = MRF:GetOption(ModOpt, "outOfCombat")
		
		oocOpt:OnUpdate(function(var)
			if var == nil then
				oocOpt:Set(true)
			elseif var then
				combatDisabler = not GameLib.GetPlayerUnit():IsInCombat()
				Apollo.RegisterEventHandler("UnitEnteredCombat", "OnCombat", MTM)
			else
				combatDisabler = false
				Apollo.RemoveEventHandler("UnitEnteredCombat", MTM)
			end
		end)
		
		function MTM:OnCombat(tUnit, bCombat)
			if tUnit:IsThePlayer() then
				combatDisabler = not bCombat
			end
		end
	end
	
	do --score units
		local scoreOpt = MRF:GetOption(ModOpt, "scoreing")
		MRF:GatherUpdates(scoreOpt)
		
		local function default(opt, val)
			opt:OnUpdate(function(x)
				if x==nil then
					opt:Set(val)
				end
			end)
			return opt
		end
		default(MRF:GetOption(scoreOpt, "maxDist"), 35)
		default(MRF:GetOption(scoreOpt, "perfDist"), 5)
		default(MRF:GetOption(scoreOpt, "offsetPlayer"), -1)
		default(MRF:GetOption(scoreOpt, "offsetFocus"), 1)
		default(MRF:GetOption(scoreOpt, "healthTotal"), 100)
		default(MRF:GetOption(scoreOpt, "healthCap"), 100)
		default(MRF:GetOption(scoreOpt, "angleTotal"), 50)
		default(MRF:GetOption(scoreOpt, "angleCap"), 47)
		default(MRF:GetOption(scoreOpt, "distTotal"), 0)
		default(MRF:GetOption(scoreOpt, "distCap"), 0)
		
		local newLine = [[ 
 ]]
		local checks = {"maxDist", "perfDist", "healthTotal", "healthCap", "angleTotal", "angleCap", "distTotal", "distCap"}
		
		--returns true for fail, false for success
		local function checkData(data)
			if not data then return true end
			for _, str in ipairs(checks) do
				if type(data[str]) ~= "number" then
					return true
				end
			end
			return false
		end
	
		scoreOpt:OnUpdate(function(data)
			if checkData(data) then return end
			local f = "function(tUnit, pVec, pFac)"
			local function add(b, yes, no)
				if b then
					f = f..newLine..yes
				elseif no then
					f = f..newLine..no
				end
			end
			
			--always needed for checking perfect/max distance:
			add(true, "local tVec = vec3new(tUnit:GetPosition() or {x=0,y=0,z=0})-pVec")
			add(true, "local dist = tVec:Length()")
			add(true, "if dist > ".. tostring(data.maxDist) .." then return nil end")
			
			local calcAng, calcDst, calcHP = data.angleTotal>0 and data.angleCap>0, data.distTotal>0 and data.distCap>0, data.healthTotal>0 and data.healthCap>0
			
			--angle and distance calculations (0-1)
			add(calcAng, "local ang = 1")
			add(calcDst, "local dst = 1")
			add(calcAng or calcDst, "if dist > ".. tostring(data.perfDist) .." then")
				add(calcAng, "ang = 1-acos(vec2dot(pFac, vec2new(tVec.x, tVec.z):Normal()))/pi")
				add(calcDst, "dst = 1-(dist- ".. tostring(data.perfDist) ..")/ ".. tostring(data.maxDist-data.perfDist))
			add(calcAng or calcDst, "end")
			
			--health calculation (0-1)
			add(calcHP, "local hp = 1-(tUnit:GetHealth() or 0) / max(tUnit:GetMaxHealth() or 0,1)")
			
			local returns = {}
			if calcAng then
				local capped = data.angleTotal > data.angleCap
				table.insert(returns, capped and "min(ang*".. tostring(data.angleTotal) ..", ".. tostring(data.angleCap) ..")" 
												or "ang*".. tostring(data.angleTotal))
			end
			if calcDst then
				local capped = data.distTotal > data.distCap
				table.insert(returns, capped and "min(dst*".. tostring(data.distTotal) ..", ".. tostring(data.distCap) ..")" 
												or "dst*".. tostring(data.distTotal))
			end
			if calcHP then
				local capped = data.healthTotal > data.healthCap
				table.insert(returns, capped and "min(hp*".. tostring(data.healthTotal) ..", ".. tostring(data.healthCap) ..")" 
												or "hp*".. tostring(data.healthTotal))
			end
			if not calcAng and not calcDst and not calcHP then
				table.insert(returns, "0") --the sum of none should be 0, not nil.
			end
			
			add(true, "return "..table.concat(returns, " + "))
			add(true, "end") --function end
			
			f = "return function(min, max, acos, pi, vec3new, vec2new, vec2dot) return "..f.." end"
			scoreUnit = loadstring(f)()(math.min, math.max, math.acos, math.pi, Vector3.New, Vector2.New, Vector2.Dot)
		end)		
	end
	
	do --target selection
		local group = {}
		local focus = nil
		local units = MRF:GetUnits()
		local offPlayer = 0
		local offFocus = 0
		
		function MTM:OnFrameCount() --this only gets triggered, if the module is active
			if combatDisabler or disabledByKey() then return end --check whether we want to change anything
			local points = nil;
			local unit = nil;
			
			local player = GameLib.GetPlayerUnit()
			local facing = player:GetFacing() or {x=1, y=0, z=0}
			
			local pVec, pFac = Vector3.New(player:GetPosition() or {x=0, y=0, z=0}), Vector2.New(facing.x, facing.z)
			
			if player and player:IsDead() == false then
				local score = scoreUnit(player, pVec, pFac)
				if score and (not points or score+offPlayer >= points) then
					unit = player
					points = score+offPlayer
				end
			end
			for idx, tUnit in pairs(group) do
				if units[idx] and tUnit:IsDead() == false then
					local score = scoreUnit(tUnit, pVec, pFac)
					if score and (not points or score >= points) then
						unit = tUnit
						points = score
					end
				else
					group[idx] = nil
				end
			end
			if focus and focus:IsDead() == false then
				local score = scoreUnit(focus, pVec, pFac)
				if score and (not points or score+offFocus >= points) then
					unit = focus
					points = score+offFocus
				end
			end
			
			if unit and unit~=target then
				GameLib.SetTargetUnit(unit)
			end
		end
		
		MRF:GetOption(ModOpt, "scoreing", "offsetPlayer"):OnUpdate(function(val)
			if type(val) == "number" then offPlayer = val end
		end)
		MRF:GetOption(ModOpt, "scoreing", "offsetFocus"):OnUpdate(function(val)
			if type(val) == "number" then offFocus = val end
		end)
		
		if MRF.RequestUpdates then --MRF v1.12a and upwards
			MRF:RequestUpdates(false, MTM)
			function MTM:miscUpdate(_, tUnit)
				local idx = tUnit:GetMemberIdx()
				if idx == 1 then return end --1=player
				group[idx] = tUnit:GetRealUnit()
			end
		else
			local mod, opt = MRF:newModule("Targeting Module", "misc", false)

			function mod:miscUpdate(_, tUnit) 
				local idx = tUnit:GetMemberIdx()
				if idx == 1 then return end --1=player
				group[idx] = tUnit:GetRealUnit()
			end
			
			opt = MRF:GetOption(opt, "activated")
			opt:OnUpdate(function(val)
				if not val then opt:Set(true) end
			end)
			opt:ForceUpdate()
			
			function mod:InitMiscSettings(parent)
				local L = MRF:Localize({
					[1] = "This is the Targeting-Modules hook into the Addon.",
					[2] = "Please visit the 'Targeting'-Tab for Options!",
					[3] = "The Checkbox on the top does nothing.",
				},{
					[1] = "Dies ist der Eingang des Targeting-Moduls in das Addon.",
					[2] = "Für Die Optionen dieses Moduls, wähle 'Targeting' in der Liste links.",
					[3] = "Der Knopf oben ist nutzlos.",
				},{})
				
				MRF:LoadForm("HalvedRow", parent):SetText(L[1])
				MRF:LoadForm("HalvedRow", parent):SetText(L[2])
				MRF:LoadForm("HalvedRow", parent):SetText(L[3])
				
				local anchor = {parent:GetAnchorOffsets()}
				anchor[4] = anchor[2] + 90
				parent:SetAnchorOffsets(unpack(anchor))
				parent:ArrangeChildrenVert()
				parent:SetSprite("BK3:UI_BK3_Holo_InsetSimple")
			end
		end
		
		Apollo.RegisterEventHandler("TargetUnitChanged", "TargetChanged", MTM)
		function MTM:TargetChanged(tUnit)
			target = tUnit
		end
		
		Apollo.RegisterEventHandler("AlternateTargetUnitChanged", "FocusChanged", MTM)
		function MTM:FocusChanged(tUnit)
			if tUnit then
				local pos = tUnit:GetDispositionTo(GameLib.GetPlayerUnit())
				if pos ~= Unit.CodeEnumDisposition.Friendly then
					tUnit = nil
				end
			end
			focus = tUnit
		end
		
		local timer = ApolloTimer.Create(5, true, "InitialFocusCheck", MTM)
		function MTM:InitialFocusCheck()
			local player = GameLib.GetPlayerUnit()
			if player then
				timer:Stop()
				timer = nil
				self:FocusChanged(player:GetAlternateTarget())
			end
		end
	end
	
	do --target drawing
		local nOff, pOff = -50, 50
		local drawOpt = MRF:GetOption(ModOpt, "drawing")
	
		function MTM:RedrawTarget()
			local pos = target and target:GetPosition()
			if pos then
				local pos = GameLib.WorldLocToScreenPoint(Vector3.New(pos))
				targetForm:SetAnchorOffsets(pos.x+nOff, pos.y+nOff, pos.x+pOff, pos.y+pOff)
				targetForm:Show(true, true)
			else
				targetForm:Show(false, true)
			end
		end
		
		MRF:GetOption(drawOpt, "activate"):OnUpdate(function(val) 
			if val == nil then
				MRF:GetOption(drawOpt, "activate"):Set(true)
			elseif val and activated() then
				Apollo.RegisterEventHandler("NextFrame", "RedrawTarget", MTM)
			else
				Apollo.RemoveEventHandler("NextFrame", MTM)
				targetForm:Show(false, true)
			end
		end)
		MRF:GetOption(ModOpt, "classes"):OnUpdate(function()
			MRF:GetOption(drawOpt, "activate"):ForceUpdate()
		end)
		
		MRF:GetOption(drawOpt, "color"):OnUpdate(function(val)
			if val == nil then
				MRF:GetOption(drawOpt, "color"):Set("FFBD2F00")
			else
				targetForm:SetBGColor(ApolloColor.new(val))
			end
		end)
		
		MRF:GetOption(drawOpt, "size"):OnUpdate(function(val)
			if val == nil then
				MRF:GetOption(drawOpt, "size"):Set(45)
			else
				nOff = -math.floor(val/2)
				pOff = val+nOff
			end
		end)
	end
	
	do --settings (UI-Components)
		local L = MRF:Localize(
			{ --English
				["txtPointsIntro1"] = "The scoring system, by which this Addon chooses targets.",
				["txtPointsIntro2"] = "Points are additive, the character achieving the most points will be targeted.",
				["ttMaxDist"] = "The maximal distance of units being targeted.",
				["ttPerfDist"] = "Assigning maximal points through angle and distance for units being closer than this.",
				["ttOffPlayer"] = "The Players points will always be offset by this value. Useful when trying to priorize others, over yourself.",
				["ttOffFocus"] = "The Points of your Focus will always be offset by this value. Useful when priorizing single targets.",
				["ttPointsTable"] = [[In this table, you define how many points a character can get from different sources.
				Note: The more points you give through a single source, the more you priorize it.]],
				["ttPointsPoints"] = [[The scaling used for the specific sources.
				0 = Deactivate this source]],
				["ttPointsCap"] = [[Apply a cap to points granted. Useful to have a broader 'perfect situation'.
				Every point above the cap will not be granted. A Cap above (or equal) to Points will affect nothing.
				Example: You grant 100 Points to angles and cap it at 50: this means everybody within the 180° infront of your character will gain 50 Points from their angle.]],
				["ttPointsHealth"] = [[Points granted for having lost health: The lower the health-percentage is, the more points one gains.]],
				["ttPointsAngle"] = [[Points granted for being infront of you. Having a low angle between your facing and the direction towards the character grants more points.]],
				["ttPointsDist"] = [[Points granted for being close to you. The closer the character is to the perfect distance, the more points he gets (being inside perfect distance grants maximal points)]],
			},
			{ --German			
				["General"] = "Allgemeines",
				["Keybinds"] = "Tastaturkürzel",
				["Points"] = "Punkte",
				["Module activated for classes:"] = "Modul aktiv für die Klassen:",
				["Disable while out of combat"] = "Nur im Kampf aktiv.",
				["Warrior"] = "Krieger",
				["Engineer"] = "Inguineur",
				["Esper"] = "Esper",
				["Medic"] = "Sanitäter",
				["Stalker"] = "Meuchler",
				["Spellslinger"] = "Arkanschütze",
				
				--keybinds
				["Stop changing targets if:"] = "Ziel nicht wechseln, falls:",
				["'Alt'-Key is held"] = "'Alt'-Taste gedrückt",
				["'Ctrl'-Key is held"] = "'Strg'-Taste gedrückt",
				["'Shift'-Key is held"] = "die Hochstelltaste (Shift) gedrückt",
				["'Caps' is:"] = "die Feststelltaste (Caps) ist:",
				["Active"] = "Eingeschalten",
				["Inactive"] = "Ausgeschalten",
				
				--points
				["txtPointsIntro1"] = "Das Punktesystem, nach welchem das Modul Ziele wählt.",
				["txtPointsIntro2"] = "Die Punkte addieren sich auf. Der Charakter mit den meisten Punkten wird ins Zeil genommen.",
				["ttMaxDist"] = "Die Distanz, ab welcher Charaktere nicht mehr angewählt werden können.",
				["ttPerfDist"] = "Für Charaktere, die näher sind, werden maximale Punkte für Winkel und Distanz vergeben.",
				["ttOffPlayer"] = "Der Spieler bekommt diese Punkte permanent angerechnet. Nützlich, wenn andere Ziele immer vor dem Spieler gewählt werden sollen.",
				["ttOffFocus"] = "Der Fokus bekommt diese Punkte permanent angerechnet. Nützlich, um einzelne Ziele besonders zu priorisieren.",
				["ttPointsTable"] = [[In dieser Tabelle wird definiert, wie viele Punkte ein Charakter aus verschiedenen Quellen erhalten kann.
				Tipp: Umso mehr Punkte durch eine Quelle vergeben werden können, umso wichtiger ist sie bei der Zielauswahl.]],
				["ttPointsPoints"] = [[Die Skalierung der jeweiligen Quelle.
				0 = Dektiviere Punkte aus dieser Quelle.]],
				["ttPointsCap"] = [[Setze die maximalen Punkte erhalten aus der Quelle. Nützlich um eine breitere 'Perfekte Situation' zu erhalten.
				Jeder Punkt über dem maximum wird nicht erhalten. Ein Maximum über (oder gleich) den Punkten wird nichts ändern.
				Beispiel: Es werden 100 Punkte mit einem Maximum von 50 durch Winkel vergeben: Das bedeutet, jeder in den 180° vor deinem Charakter wird die gleichen 50 Punkte durch Winkel erhalten.]],
				["ttPointsHealth"] = [[Punkte für verlorenes Leben: Geringeres prozentuales Leben resultiert in mehr Punkten.]],
				["ttPointsAngle"] = [[Punkte für Positionierung vor dem Spieler. Ein kleiner Winkel zwischen der Blickrichtung des Spielers und Richtung zu dem Charakter erbringt mehr Punkte.]],
				["ttPointsDist"] = [[Punkte für Nähe. Umso näher der Charakter an der perfekten Distanz ist, umso mehr Punkte erhält er. (Innerhlab der perfekten Distanz wird immer das maximum vergeben)]],
				["Maximal Distance"] = "Maximale Distanz",
				["Perfect Distance"] = "Perfekte Distanz",
				["Offset - Player:"] = "Versatz - Spieler:",
				["Offset - Focus:"] = "Versatz - Fokus:",
				["Category"] = "Kategorie",
				["Health"] = "Leben",
				["Angle"] = "Winkel",
				["Distance"] = "Distanz",
				["Cap"] = "Maximum",
				
				--drawing
				["Target Marker:"] = "Zielmarkierung:",
				["Activate Target Marker"] = "Aktiviere Markierung",
				["Target Color:"] = "Markierungsfarbe:",
				["Target Size:"] = "Markierungsgröße:",
			},
			{ --French
			}
		)
		
		MRF:AddMainTab("Targeting") --could actually localize 'Targeting' aswell - do we want to?
		MRF:AddChildTab(L["General"], "Targeting", MTM, "GeneralSettings")
		MRF:AddChildTab(L["Keybinds"], "Targeting", MTM, "KeybindSettings")
		MRF:AddChildTab(L["Points"], "Targeting", MTM, "PointsSettings")
		
		local function sizeParent(wnd)
			wnd:SetSprite("BK3:UI_BK3_Holo_InsetSimple")
			wnd:ArrangeChildrenVert()
			local children = wnd:GetChildren()
			local l,t,r,b = wnd:GetAnchorOffsets()
			wnd:SetAnchorOffsets(l,t,r,select(4,children[#children]:GetAnchorOffsets())+t)
			wnd:GetParent():RecalculateContentExtents()
		end
		
		local function prepParent(parent, name)
			local parent = MRF:LoadForm("SimpleTab", parent)
			parent:FindChild("Title"):SetText("Targeting - "..name)
			return parent:FindChild("Space")
		end
		
		function MTM:GeneralSettings(...)
			local parent = prepParent(...)
			
			MRF:LoadForm("HalvedRow", parent):SetText(L["Module activated for classes:"])
			
			local row = MRF:LoadForm("HalvedRow", parent)
			MRF:applyCheckbox(row:FindChild("Left"), MRF:GetOption(ModOpt, "classes", GameLib.CodeEnumClass.Warrior), L["Warrior"])
			MRF:applyCheckbox(row:FindChild("Right"), MRF:GetOption(ModOpt, "classes", GameLib.CodeEnumClass.Engineer), L["Engineer"])
			
			row = MRF:LoadForm("HalvedRow", parent)
			MRF:applyCheckbox(row:FindChild("Left"), MRF:GetOption(ModOpt, "classes", GameLib.CodeEnumClass.Esper), L["Esper"])
			MRF:applyCheckbox(row:FindChild("Right"), MRF:GetOption(ModOpt, "classes", GameLib.CodeEnumClass.Medic), L["Medic"])
			
			row = MRF:LoadForm("HalvedRow", parent)
			MRF:applyCheckbox(row:FindChild("Left"), MRF:GetOption(ModOpt, "classes", GameLib.CodeEnumClass.Stalker), L["Stalker"])
			MRF:applyCheckbox(row:FindChild("Right"), MRF:GetOption(ModOpt, "classes", GameLib.CodeEnumClass.Spellslinger), L["Spellslinger"])
			
			MRF:LoadForm("HalvedRow", parent)
			
			row = MRF:LoadForm("HalvedRow", parent)
			MRF:applyCheckbox(row:FindChild("Left"), MRF:GetOption(ModOpt, "outOfCombat"), L["Disable while out of combat"])
			
			MRF:LoadForm("HalvedRow", parent):SetText(L["Target Marker:"])
			row = MRF:LoadForm("HalvedRow", parent)
			MRF:applyCheckbox(row:FindChild("Left"), MRF:GetOption(ModOpt, "drawing", "activate"), L["Activate Target Marker"])
			local target = MTM:LoadForm("Target", row:FindChild("Right"))
			target:SetAnchorOffsets(-15,-15,15,15)
			target:SetAnchorPoints(0.5,0.5,0.5,0.5)
			MRF:GetOption(ModOpt, "drawing", "color"):OnUpdate(function(val)
				if not val then return end
				target:SetBGColor(ApolloColor.new(val))
			end)
			
			row = MRF:LoadForm("HalvedRow", parent)
			row:FindChild("Left"):SetText(L["Target Color:"])
			MRF:applyColorbutton(row:FindChild("Right"), MRF:GetOption(ModOpt, "drawing", "color"))
			
			row = MRF:LoadForm("HalvedRow", parent)
			row:FindChild("Left"):SetText(L["Target Size:"])
			MRF:applySlider(row:FindChild("Right"), MRF:GetOption(ModOpt, "drawing", "size"), 1, 100, 1, false, false, true)
			
			sizeParent(parent)
			ModOpt:ForceUpdate()
		end
		
		function MTM:KeybindSettings(...)
			local parent = prepParent(...)
			
			MRF:LoadForm("HalvedRow", parent):SetText(L["Stop changing targets if:"])
			
			local row = MRF:LoadForm("HalvedRow", parent)
			MRF:applyCheckbox(row:FindChild("Left"), MRF:GetOption(ModOpt, "disableKeys", "alt"), L["'Alt'-Key is held"])
			
			row = MRF:LoadForm("HalvedRow", parent)
			MRF:applyCheckbox(row:FindChild("Left"), MRF:GetOption(ModOpt, "disableKeys", "ctrl"), L["'Ctrl'-Key is held"])
			
			row = MRF:LoadForm("HalvedRow", parent)
			MRF:applyCheckbox(row:FindChild("Left"), MRF:GetOption(ModOpt, "disableKeys", "shift"), L["'Shift'-Key is held"])
			
			row = MRF:LoadForm("HalvedRow", parent)
			MRF:applyCheckbox(row:FindChild("Left"), MRF:GetOption(ModOpt, "disableKeys", "caps"), L["'Caps' is:"])
			MRF:applyDropdown(row:FindChild("Right"), {true, false}, MRF:GetOption(ModOpt, "disableKeys", "capsOnOff"), {[true] = L["Active"], [false] = L["Inactive"]})
			
			
			sizeParent(parent)
			ModOpt:ForceUpdate()
		end
		
		function MTM:PointsSettings(...)
			local parent = prepParent(...)
			local row;
			
			MRF:LoadForm("HalvedRow", parent):SetText(L["txtPointsIntro1"]) -- The Scoring System, by which this Addon chooses targets. 
			MRF:LoadForm("HalvedRow", parent):SetText(L["txtPointsIntro2"]) --Points are Additive, the Character achieving the most points will be targeted.
			MRF:LoadForm("HalvedRow", parent)
			
			--max distance --the maximal distance of units being targeted
			row = MRF:LoadForm("HalvedRow", parent)
			MRF:LoadForm("QuestionMark", row:FindChild("Left")):SetTooltip(L["ttMaxDist"])
			row:FindChild("Left"):SetText(L["Maximal Distance:"])
			MRF:applySlider(row:FindChild("Right"), MRF:GetOption(ModOpt, "scoreing", "maxDist"), 1, 100, 1, true, false, true)
			
			--perfect distance --assuming perfect radius for units being closer than this distance
			row = MRF:LoadForm("HalvedRow", parent)
			MRF:LoadForm("QuestionMark", row:FindChild("Left")):SetTooltip(L["ttPerfDist"])
			row:FindChild("Left"):SetText(L["Perfect Distance:"])
			MRF:applySlider(row:FindChild("Right"), MRF:GetOption(ModOpt, "scoreing", "perfDist"), 1, 100, 1, true, false, true)
			
			--offset: player
			row = MRF:LoadForm("HalvedRow", parent)
			MRF:LoadForm("QuestionMark", row:FindChild("Left")):SetTooltip(L["ttOffPlayer"])
			row:FindChild("Left"):SetText(L["Offset - Player:"])
			MRF:applySlider(row:FindChild("Right"), MRF:GetOption(ModOpt, "scoreing", "offsetPlayer"), -50, 50, 1, false, true)
			
			--offset: focus
			row = MRF:LoadForm("HalvedRow", parent)
			MRF:LoadForm("QuestionMark", row:FindChild("Left")):SetTooltip(L["ttOffFocus"])
			row:FindChild("Left"):SetText(L["Offset - Focus:"])
			MRF:applySlider(row:FindChild("Right"), MRF:GetOption(ModOpt, "scoreing", "offsetFocus"), -50, 50, 1, false, true)
			
			MRF:LoadForm("HalvedRow", parent)
			--TableRow--Left --Center --Right
			--Main Categorys: Headers
			row = self:LoadForm("TableRow", parent)
			MRF:LoadForm("QuestionMark", row:FindChild("Left")):SetTooltip(L["ttPointsTable"])
			row:FindChild("Left"):SetText(L["Category"])
			MRF:LoadForm("QuestionMark", row:FindChild("Center")):SetTooltip(L["ttPointsPoints"])
			row:FindChild("Center"):SetText(L["Points"])
			MRF:LoadForm("QuestionMark", row:FindChild("Right")):SetTooltip(L["ttPointsCap"])
			row:FindChild("Right"):SetText(L["Cap"])
			
			--Main Categorys: Health
			row = self:LoadForm("TableRow", parent)
			MRF:LoadForm("QuestionMark", row:FindChild("Left")):SetTooltip(L["ttPointsHealth"])
			row:FindChild("Left"):SetText(L["Health"])
			MRF:applySlider(row:FindChild("Center"), MRF:GetOption(ModOpt, "scoreing", "healthTotal"), 0, 100, 1, false, false, true)
			MRF:applySlider(row:FindChild("Right"), MRF:GetOption(ModOpt, "scoreing", "healthCap"), 0, 100, 1, false, false, true)
			
			--Main Categorys: Angle
			row = self:LoadForm("TableRow", parent)
			MRF:LoadForm("QuestionMark", row:FindChild("Left")):SetTooltip(L["ttPointsAngle"])
			row:FindChild("Left"):SetText(L["Angle"])
			MRF:applySlider(row:FindChild("Center"), MRF:GetOption(ModOpt, "scoreing", "angleTotal"), 0, 100, 1, false, false, true)
			MRF:applySlider(row:FindChild("Right"), MRF:GetOption(ModOpt, "scoreing", "angleCap"), 0, 100, 1, false, false, true)
			
			--Main Categorys: Distance
			row = self:LoadForm("TableRow", parent)
			MRF:LoadForm("QuestionMark", row:FindChild("Left")):SetTooltip(L["ttPointsDist"])
			row:FindChild("Left"):SetText(L["Distance"])
			MRF:applySlider(row:FindChild("Center"), MRF:GetOption(ModOpt, "scoreing", "distTotal"), 0, 100, 1, false, false, true)
			MRF:applySlider(row:FindChild("Right"), MRF:GetOption(ModOpt, "scoreing", "distCap"), 0, 100, 1, false, false, true)
			
			sizeParent(parent)
			ModOpt:ForceUpdate()
		end
		
	end
		
	ModOpt:ForceUpdate() --at this point the saves of mrf are already loaded. We will not get an automatic Update from MRF's SaveHandler; so we do it ourselves
end
-----------------------------------------------------------------------------------------------
-- MTM Instance
-----------------------------------------------------------------------------------------------
local MTMInst = MTM:new()
MTMInst:Init()
